﻿/*
# #############################################################################
#  ATTENTION
# #############################################################################
#
#  This software is for guidance only. It is provided "as is".
#  It only aims at providing coding information to the customer in order for
#  them to save time. As a result no warranties, whether express, implied or
#  statutory, including, but not limited to implied warranties of
#  merchantability and fitness for a particular purpose apply to this software.
#  A·P·E Angewandte Physik und Elektronik GmbH shall not be held liable for any
#  direct, indirect or consequential damages with respect to any claims arising
#  from the content of such sourcecode and/or the use made by customers of the
#  coding information contained herein in connection with their products.
#
#  File Version: 1.0
#
#  Changelog:
#    1.0 - Initial Version
#
# #############################################################################
*/

using System;
using System.Net;
using System.Net.Sockets;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.IO;

namespace APE_Interface_Example
{
    class APE_Device
    {
        private string remote_hostname;
        private int remote_port;
        private string devicename;
        private bool status_connected;
        private IPHostEntry ipHostInfo;
        private IPAddress ipAddress;
        private IPEndPoint remoteEP;
        private Socket remote_device;

        private const int BUFFER_LEN = 524288; //Buffer length for receive buffer (512KB)

        //---------------------------------------------------------------------
        //Class Constructor
        //
        //Creates a new instance of an ape_device object and tries to connect
        //to the given remote target
        //
        //Parameters:
        //  host - string - Hostname or IP Address of remote target
        //  port - int32  - TCP Port of remote target
        //  name - string - a name for the device
        //
        //Returns:
        //  none
        //
        //---------------------------------------------------------------------
        public APE_Device (string host, int port, string name = "APEdevice")
        {
            bool connected = false;
            bool retry = true;

            if (host.Length == 0)
            {
                Console.WriteLine("Hostname must not be empty!");
                return;
            }
            remote_hostname = host;

            if (port < 0 ||port > 65535)
            {
                Console.WriteLine("Portnumber must be within range of 1 .. 65535!");
                return;
            }
            remote_port = port;

            if (name.Length != 0)
                devicename = name;

            while(!connected && retry)
            {
               connected = open_connection();

               if(!connected)
               {
                   Console.WriteLine("Connect falied. Retry? (Y/n)");
                   retry = (Console.ReadLine()[0] != 'n');
               }
            }
            status_connected = connected;
        }

        //---------------------------------------------------------------------
        //Set_hostname
        //
        //Set a new hostname for the connection. If a connection is still active
        //it is closed first. After setting the hostname call open_connection()
        //to reconnect.
        //
        //Parameters:
        //  host - string - Hostname or IP Address of remote target
        //
        //Returns:
        //  hostname - string - new value of hostname
        //
        //---------------------------------------------------------------------
        public string set_hostname(string host)
        {
            if (is_connected())
            {
                disconnect();
            }

            if (host.Length == 0)
            {
                Console.WriteLine("Hostname must not be empty!");
                return "";
            }
            remote_hostname = host;

            return remote_hostname;
        }

        //---------------------------------------------------------------------
        //Set_port
        //
        //Set a new tcp port for the connection. If a connection is still active
        //it is closed first. After setting the hostname call open_connection()
        //to reconnect.
        //
        //Parameters:
        //  port - int32 - TCP Port of remote target
        //
        //Returns:
        //  port - int32 - new Portnumber
        //
        //---------------------------------------------------------------------
        public int set_port(int port)
        {
            if (is_connected())
            {
                disconnect();
            }

            if (port < 0 ||port > 65535)
            {
                Console.WriteLine("Portnumber must be within range of 1 .. 65535!");
                return -1;
            }
            remote_port = port;

            return remote_port;
        }

        //---------------------------------------------------------------------
        //Set_devicename
        //
        //Set a new devicename. This name can be used to identify multiple
        //instances of ape_devices. it is not used for communication.
        //
        //Parameters:
        //  name - string - New name for the ape-device object
        //
        //Returns:
        //  name - string - new value of devicename
        //
        //---------------------------------------------------------------------
        public string set_devicename(string name) 
        {
            if (name.Length != 0)
            {
                devicename = name;
                return devicename;
            }
            return "";
        }

        //---------------------------------------------------------------------
        //Open_connection
        //
        //Tries to open a socket connection to the set host/port. 
        //The connection details are set by the class constructor or can be 
        //set manually by set_hostname() and set_port()
        //
        //Parameters:
        //  none
        //
        //Returns:
        //  status - bool - Connection Status (true if socket connection was
        //                  opened, false if an error occurred)
        //
        //---------------------------------------------------------------------
        public bool open_connection()
        {
            try
            {
                IPAddress[] ipv4Addresses = Array.FindAll(Dns.GetHostAdresses(remote_hostname), a => a.AddressFamily == AddressFamily.InterNetwork);
                ipAddress = ipv4Addresses[0];
                remoteEP = new IPEndPoint(ipAddress, remote_port);

                // Create a TCP/IP  socket.
                remote_device = new Socket(AddressFamily.InterNetwork, 
                                    SocketType.Stream, 
                                    ProtocolType.Tcp
                                    );

                // Connect the socket to the remote endpoint. Catch any errors.
                try
                {
                    remote_device.Connect(remoteEP);
                    Console.WriteLine(  "[Connect] Socket connected to {0}", 
                                        remote_device.RemoteEndPoint.ToString());

                    //Wait 0.5 Seconds before continuing
                    Thread.Sleep(500);

                    return true;
                }
                catch (ArgumentNullException ane)
                {
                    Console.WriteLine(  "ArgumentNullException : {0}", 
                                        ane.ToString());
                }
                catch (SocketException se)
                {
                    Console.WriteLine(  "SocketException : {0}", 
                                        se.ToString());
                }
                catch (Exception e)
                {
                    Console.WriteLine(  "Unexpected exception : {0}", 
                                        e.ToString());
                }
            } 
            catch (Exception e) 
            {
                Console.WriteLine( e.ToString());
            }
            
            return false;
        }

        //---------------------------------------------------------------------
        //Disconnect
        //
        //Close the socket connection, if it is still opened
        //
        //Parameters:
        //  none
        //
        //Returns:
        //  status - bool - Connection Status (true if socket connection was
        //                  closed, false if connection was already closed)
        //
        //---------------------------------------------------------------------
        public bool disconnect()
        {
            if (is_connected())
            {
                // Release the socket.
                remote_device.Shutdown(SocketShutdown.Both);
                remote_device.Close();
                return true;
            }
            return false;
        }

        //---------------------------------------------------------------------
        //is_connected
        //
        //Returns the current status of the connection
        //
        //Parameters:
        //  none
        //
        //Returns:
        //  status - bool - true if socket connection is open, else false
        //
        //---------------------------------------------------------------------
        public bool is_connected()
        {
            return status_connected;
        }

        //---------------------------------------------------------------------
        //recv
        //
        //Try to read data from the socket connection
        //
        //Parameters:
        //  len - int - Number of characters to read from the socket
        //  read_scpi_block - bool - Status of SCPI Decoding option
        //
        //Returns:
        //  data - string - Data read from the socket as c# string
        //
        //---------------------------------------------------------------------
        public string recv(int len, bool read_scpi_block = false)
        {
            int bytesRec = 0;
            int read_len = 0;
            string read_buffer = "";
            byte[] recv_buffer_b;

            if (!is_connected())
            {
                Console.WriteLine("[RECV] Error. Not connected");
                return "";
            }
            if (len == 0)
                return "";

            if (read_scpi_block)
            {
                return receive_scpi();
            }

            //If length < 0 create full 512KB buffer
            if (len < 0) 
            {
                recv_buffer_b = new byte[BUFFER_LEN];
                read_len = BUFFER_LEN;
            }
            else
            {
                recv_buffer_b = new byte[len];
                read_len = len;
            }

            try
            {
                read_buffer = "";
                //Get 8 Bit Encoding table (ASCII Encoding uses only 7 Bits, which results in data loss)
                Encoding enc = Encoding.GetEncoding(1252); 

                // Receive data from the remote device and convert it to a string object
                do
                {
                    bytesRec = remote_device.Receive(recv_buffer_b, 0, read_len, SocketFlags.None);
                    if (bytesRec > 0)
                    {
                        read_buffer += enc.GetString(recv_buffer_b, 0, bytesRec);

                        read_len -= bytesRec;
                    }
                } while ((bytesRec > 0) && (read_len > 0) && (len > 0));
            }
            catch (ArgumentNullException ane)
            {
                Console.WriteLine("ArgumentNullException : {0}", ane.ToString());
                return "";
            }
            catch (SocketException se)
            {
                Console.WriteLine("SocketException : {0}", se.ToString());
                return "";
            }
            catch (Exception e)
            {
                Console.WriteLine("Unexpected exception : {0}", e.ToString());
                return "";
            }
  
            return read_buffer;
        }

        //---------------------------------------------------------------------
        //receive_scpi()
        //
        //Try to read an SCPI Block from the socket connection. 
        //The function automatically decodes the block and returns the pure
        //data content
        //
        //Parameters:
        //  none
        //
        //Returns:
        //  data - string - Received data as c# string
        //
        //---------------------------------------------------------------------
        private string receive_scpi()
        {
            int header_len = 0;
            int data_len = 0;
            string read_buffer = "";

            //Read first character, this must be a # if data is an scpi block
            read_buffer = recv(1);
            if (read_buffer != "#")
            {
                Console.WriteLine("[READ SCPI] Error. Expected \"#\" as first character, but got \"" + read_buffer[0] + "\"");
                return "";
            }
            read_buffer = "";
            //Read second character, this must be a numeric character, that reperesents the length of the header
            read_buffer = recv(1);
            if (!read_buffer.All(Char.IsDigit))
            {
                Console.WriteLine("[READ SCPI] Error. Expected a number as second character, but got\"" + read_buffer[0] + "\"");
                return "";
            }
            //Convert read number into an integer value
            header_len = System.Convert.ToInt32(read_buffer);
            if (header_len <= 0)
            {
                Console.WriteLine("[READ SCPI] Error. Header Length was less or qeual to Zero.");
                return "";
            }
            read_buffer = "";

            //Read number of characters, defined in header_len
            read_buffer = recv(header_len);
            if (!read_buffer.All(Char.IsDigit))
            {
                Console.WriteLine("[READ SCPI] Error. Expected " + header_len.ToString() + " digits, but got \"" + read_buffer + "\"");
                return "";
            }
            //Convert read number into an integer value and check, that it is greater than zero
            data_len = System.Convert.ToInt32(read_buffer);
            if (header_len <= 0)
            {
                return "";
            }
            //Receive scpi data content
            read_buffer = recv(data_len);

            //Return received data
            return read_buffer;
        }

        //---------------------------------------------------------------------
        //send
        //
        //Send a command-string via the socket connection
        //
        //Parameters:
        //  command - string -  Command to send. Termination characters \r\n
        //                      are automatically appended
        //
        //Returns:
        //  bytes - int - Number of bytes/characters sent
        //
        //---------------------------------------------------------------------
        public int send(string command)
        {
            if (command.Length != 0)
            {
                try
                { 
                    //Convert the command string into a byte array.
                    byte[] msg = Encoding.ASCII.GetBytes(command+"\r\n" );

                    // Send the byte array through the socket.
                    int bytesSent = remote_device.Send(msg);

                    return bytesSent;
                }
                catch (ArgumentNullException ane)
                {
                    Console.WriteLine("ArgumentNullException : {0}", ane.ToString());
                    return 0;
                }
                catch (SocketException se)
                {
                    Console.WriteLine("SocketException : {0}", se.ToString());
                    return 0;
                }
                catch (Exception e)
                {
                    Console.WriteLine("Unexpected exception : {0}", e.ToString());
                    return 0;
                }
            }
            return 0;
        }

        //---------------------------------------------------------------------
        //query
        //
        //Combined send/recv function. Sends a command and tries to read an
        //response from the socket connection
        //
        //Parameters:
        //  command - string -  Command to send. Termination characters \r\n
        //                      are automatically appended
        //  read_scpi_block - bool - Status of SCPI Readout option
        //
        //Returns:
        //  data - string - Received data as c# string
        //
        //---------------------------------------------------------------------
        public string query (string command, bool read_scpi_block = false)
        {
            if (command.Length != 0)
            {
                send(command);
                return recv(-1, read_scpi_block);
            }
            return "";
        }

        //---------------------------------------------------------------------
        //get_idn
        //
        //Returns the SCPI device identification string
        //
        //Parameters:
        //  none
        //
        //Returns:
        //  idn - string - device identification string
        //
        //---------------------------------------------------------------------
        public string get_idn()
        {
            return query("*IDN?");
        }

        //---------------------------------------------------------------------
        //get_stb
        //
        //Returns the SCPI devices' status byte
        //
        //Parameters:
        //  none
        //
        //Returns:
        //  stb - string - device status byte (decimal number as string)
        //
        //---------------------------------------------------------------------
        public string get_status()
        {
            return query("*STB?");
        }

        //---------------------------------------------------------------------
        //get_oper
        //
        //Returns the SCPI devices' operation status register
        //
        //Parameters:
        //  none
        //
        //Returns:
        //  oper - string - device operation status register 
        //                  (decimal number as string)
        //
        //---------------------------------------------------------------------
        public string get_oper()
        {
            return query("*OPER?");
        }
    }
}
